/////////////////////////////////////////////////////
//Doppelpendel ohne Reibung mit Runge-Kutta 4
////////////////////////////////////////////////////

// Attribute:
int n = 4;            // Anzahl der Variablen
float[] ZVaktuell = {3.14, 3.14, 0, 0}; // Initialisierung w1,w2,g1,g2
boolean spurZeichnen;
float w1,w2; //Winkel von der Senkrechten weg
float W1,W2; //Für Runge-Kutta verwendet: Winkel in Ruhelage: PI
float g1,g2; //Winkelgeschwindigkeiten
float f = 500;        
float dt = 0.1;      // Zeitschritt
//Lage der Massen
float masse1X ;
float masse1Y ;
float masse2X ;
float masse2Y ;
float masse;
float E; //Gesamtenergie mit Nullniveau in Ruhelage
float L = 180;
float G = 9.81;         // Ortfaktor für Schwerkraft
float ruhelage = 400;         // Ruhelag in x-Koordinate
ArrayList<PVector> punkte = new ArrayList<PVector>(); //Zum speichern der Kurvenpunkte
float EKontrolleA;//Gesamtenergie 


void setup() {
  spurZeichnen = false;
  size(800,800);
  frameRate(f);
  masse =1;
  spurZeichnen = true;
  random(0.2,0.2);   
  W1 = PI-w1;
  w2= random(-0.2,0.2);
  W2 = PI-w2;
  g2= 0.1; 
  E= 1000;
  ZVaktuell[0] = W1;
  ZVaktuell[1] = W2;         
  ZVaktuell[3] = g2; 
  g1 = berechneG1();//Gesamtenergie E konstant. Wenn bekannt, kann g1 aus den anderen Parametern berechnet werden
  ZVaktuell[2] = g1;
  EKontrolleA = masse*g1*g1*L*L +0.5*masse*g2*g2*L*L + masse*g1*g2*L*L*cos(w2-w1) - G*masse*L*(2*cos(w1)+cos(w2)+3);
  //println("setup "+EKontrolleA,w1,w2,g1,g2);
 
}
/////////////////////////////////////////////////////////
void draw() { 
  pendelDisplay();
  pendelUpdate();
}
//////////////////////////////////////////////////////////
float[] ableitungZeit(float[] Y) {
    float[] ZustandsVektor = new float[n]; 
    ZustandsVektor[0]=Y[2];
    ZustandsVektor[1]=Y[3];
    // benötigte Werte :
    float sinDiff = sin(Y[0]-Y[1]);
    float cosDiff = cos(Y[0]-Y[1]);
    float cosProdukt = cosDiff * cosDiff;
    float s0 = sin(Y[0]);
    float s1 = sin(Y[1]);
    float alpha = - Y[3]*Y[3]*sinDiff + 2*G/L*s0 ;
    float beta = Y[2]*Y[2]*sinDiff + G/L*s1 ;
    ZustandsVektor[2]= 1 / (2-cosProdukt) * (alpha - cosDiff * beta); //Geschwindigkeit 1
    ZustandsVektor[3]= 1 / (2-cosProdukt) * (2*beta - cosDiff * alpha);//Geschwindigkeit 2 
    return ZustandsVektor;
}
////////////////////////////////////////////////////////////////////////////
void pendelUpdate(){
    // Zustand nach nächstem Zeitschritt mit Runge-Kutta 4 berechnen
    // "float[] ableitungZeit" entspricht der zeitlichen Ableitung dy/dt (4-dimensional). 
    
    //Die folgende Zeile ist Bestandteil für das Zeichnen eines Poincaré-Schnitts
    //float wVor = ZVaktuell[1];
    float[] K1 = ableitungZeit(ZVaktuell); 
    // K2 =ZustandsVektor(t0 + dt/2, ZVaktuell + dt/2 * K1)
    float[] ZVtemp = new float[n];
    for (int i=0; i<n; i++) {
      ZVtemp[i] = ZVaktuell[i] + K1[i]*dt*0.5;    
    }
    float[] K2 = ableitungZeit(ZVtemp);
    // K3 =ZustandsVektor(t0 + dt/2, ZVaktuell + dt/2 * K2)
    for (int i=0; i<n; i++) {
      ZVtemp[i] = ZVaktuell[i] + K2[i]*dt*0.5;
    }
    float[] K3 = ableitungZeit( ZVtemp);
    // K4 =ZustandsVektor(t0 + dt, ZVaktuell + dt * K3)
    for (int i=0; i<n; i++) {
      ZVtemp[i] = ZVaktuell[i] + K3[i]*dt;
    }
    float[] K4 = ableitungZeit( ZVtemp);
    // Zusammenfassung
    //  ZVaktuell = ZVaktuell + dt/6 * (K1 + 2*(K2+K3) + K4)
    for (int i=0; i<n; i++) {
      ZVaktuell[i] = ZVaktuell[i] + dt / 6.0 * (K1[i] + 2.0*(K2[i] + K3[i]) + K4[i]);
    }  
    //Die folgenden Zeilen sind Bestandteil für das Zeichnen eines Poincaré-Schnitts
    //float wNach = ZVaktuell[1];
    // if(wVor < PI && wNach > PI ) println(wVor,wNach,ZVaktuell[0] ,ZVaktuell[2] );
    
    //Winkel immer zwischen 0 und 2*PI 
    if(ZVaktuell[0] > 2*PI) ZVaktuell[0] = ZVaktuell[0] - 2*PI;
    if(ZVaktuell[0] < 0)    ZVaktuell[0] = ZVaktuell[0] + 2*PI;
    if(ZVaktuell[1] > 2*PI) ZVaktuell[1] = ZVaktuell[1] - 2*PI;
    if(ZVaktuell[1] < 0)    ZVaktuell[1] = ZVaktuell[1] + 2*PI;
 
    
    W1 = ZVaktuell[0]; w1 = PI-W1;
    W2 = ZVaktuell[1]; w2 = PI-W2;
    g1= ZVaktuell[2];
    g2 = ZVaktuell[3]; 
    EKontrolleA = masse*g1*g1*L*L +0.5*masse*g2*g2*L*L + masse*g1*g2*L*L*cos(w2-w1) - G*masse*L*(2*cos(w1)+cos(w2)-3);
    //println("Dauer "+EKontrolleA,w1,w2,g1,g2);//Zur Überwachung der Gesamtenergie
}    
/////////////////////////////////////////////////////////////////////////////////
void pendelDisplay(){
    background(255);
    //Positionen berechnen
    masse1X = ruhelage + L * sin( ZVaktuell[0] );
    masse1Y = ruhelage - L * cos( ZVaktuell[0] );
    masse2X = masse1X  + L * sin( ZVaktuell[1] );
    masse2Y = masse1Y  - L * cos( ZVaktuell[1] );
    if(spurZeichnen) punkte.add(new PVector(masse2X,masse2Y));
    // Pendel zeichnen
    stroke(10);
    strokeWeight(15);
    line( ruhelage, ruhelage, masse1X, masse1Y ); //Oberer Pendelstab
    line( masse1X, masse1Y, masse2X, masse2Y ); //Unterer Pendelstab
    strokeWeight(2);
    fill(0,200,0);
    ellipse( 400, 400, 15, 15 );
    // 1.Masse:
    fill(0,0,200);
    ellipse( masse1X, masse1Y, 50, 50 );//oberer Pendelkörper
    // 2.Masse:
    fill(200,0,0);
    ellipse( masse2X, masse2Y, 50, 50 ); //unterer Pendelkörper 
    if(spurZeichnen){
      strokeWeight(2);//Zurückgelegter Weg des unteren Pendels mit Farbverlauf zeichnen
      for(int i =0; i< punkte.size()-1 ;i++){
         if( i%512<256){ stroke(255-i%256,0,i%256);}  
         else{ stroke(i%256,0,255-i%256);}
        
         line(punkte.get(i).x,punkte.get(i).y,punkte.get(i+1).x,punkte.get(i+1).y);
      }
    }
}
//////////////////////////////////////////////////////////////////////////////

float berechneG1(){

  g1 = -g2*cos(w1-w2)/2 + sqrt(g2*g2*cos(w2-w1)*cos(w2-w1)/4 -g2*g2/2+ G*(2*cos(w1)+cos(w2)-3)/L + E/(L*L*masse));  
  //g1B = -g2*cos(W1-W2)/2 - sqrt(g2*g2*cos(W2-W1)*cos(W2-W1)/4 -g2*g2/2+ G*(2*cos(W1)+cos(W2)-3)/L + E/(L*L*masse));
  //println("berechneG1 "+E,w1,w2,g1,g2);
  return g1; 
}


////////////////////////////////////////////////////////////////////////////////
void mouseClicked() {//Neustart abwechselnd mit und ohne Spur
  w1=random(PI/4,3*PI/4);   
  W1 = PI-w1;
  w2=random(PI/4,3*PI/4);
  W2 = PI-w2;
  g1=random(-0.1,0.1);
  g2=random(-0.1,0.1); 
  ZVaktuell[0] = W1;
  ZVaktuell[1] = W2;         
  ZVaktuell[2] = g1;
  ZVaktuell[3] = g2; 
  punkte.clear();
  spurZeichnen = !spurZeichnen; //abwechselnd 
}
////////////////////////////////////////////////////////////////////////////////
void keyPressed() {//Animation schneller/langsamer
  if (key == '+') {
    dt = dt*1.2;
  }
  if (key == '-') {
    dt = dt/1.2;
  }
}